# Calculation of Roche lobe and stream from L1,
# based on lobe.f presented in Spruit 1998.
# Translated from Fortran to R by M. Uemura, 13.Apr.2015

gg <- 6.667e-8
msun <- 1.989e33

rlq1 <- function(q) {
  if (abs(1-q) < 1e-4) {
    return(0.5)
  } else {
    rl <- 0.0
    rn <- 1-q
    while (abs(rl/rn-1) > 1e-4) {
      rl <- rn
      f <- q/(1-rl)^2-1/rl^2+(1+q)*rl-q
      fa <- 2*q/(1-rl)^3+2/rl^3+(1+q)
      rn <- rl-f/fa
    }
    return(rn)
  }
}

pot <- function(q,x,y,z){
  r <- sqrt(x*x+y*y+z*z)
  if (abs(r)<1e-10) {
    stop("ERROR: r=0 in function pot\n")
  }
  rh <- sqrt(x*x+y*y)
  st <- rh/r
  if (abs(rh)<1e-10) {
    cf <- 1.0
  } else{
    cf <- x/rh
  }
  r2 <- 1/(1+q)
  r1 <- sqrt(1+r^2-2*r*cf*st)
  pot=-1/r-1/q/r1-0.5*(1/q+1)*(r2^2+(r*st)^2-2*r2*r*cf*st)
  pr <- 1/r^2+1/q/(r1^3)*(r-cf*st)-0.5*(1/q+1)*2*(r*st*st-r2*cf*st)
  return(list(pot=pot,pr=pr))
}

surf <- function(q,rl1,nc){
  dc <- pi/nc
  ch <- ((1:nc)-1)*pi/(nc-1)
  rs1 <- 1-rl1
  fs <- pot(q,rs1,0.0,0.0)
  rx <- (1-dc)*rs1
  r <- rs1
  x <- cos(ch)
  x <- x[-1]
  y <- sin(ch)
  y <- y[-1]
  im <- 20
  for (i in 1:(nc-1)) {
    j <- 0
    f <- fs
    f$pot <- 1.0
    while (j<im & (abs(f$pot-fs$pot)>1e-4 | j==0)){
      j <- j+1
      r1 <- rx
      f <- pot(q,r1*x[i],r1*y[i],0.0)
      rx <- r1-(f$pot-fs$pot)/f$pr
      if (rx > rs1) {rx <- rs1}
    }
    r <- c(r,rx)
  }
  return(list(r=r,ch=ch))
}

lobes <- function(q,rs,nc=72){
  l1 <- surf(q,rs,nc)
  x1 <- 1-l1$r*cos(l1$ch)
  y1 <- -l1$r*sin(l1$ch)
  return(list(x=c(x1,rev(x1)),y=c(y1,rev(-y1))))
}

intrk <- function(z,w,dt,z1,z2,qm){
  zx <- z
  wx <- w
  tmp <- eqmot(zx,wx,z1,z2,qm)
  hz0 <- tmp$zp*dt
  hw0 <- tmp$wp*dt
  zx <- z+hz0/2
  wx <- w+hw0/2
  tmp <- eqmot(zx,wx,z1,z2,qm)
  hz1 <- tmp$zp*dt
  hw1 <- tmp$wp*dt
  zx <- z+hz1/2
  wx <- w+hw1/2
  tmp <- eqmot(zx,wx,z1,z2,qm)
  hz2 <- tmp$zp*dt
  hw2 <- tmp$wp*dt
  zx <- z+hz2
  wx <- w+hw2
  tmp <- eqmot(zx,wx,z1,z2,qm)
  hz3 <- tmp$zp*dt
  hw3 <- tmp$wp*dt
  dz <- (hz0+2*hz1+2*hz2+hz3)/6
  dw <- (hw0+2*hw1+2*hw2+hw3)/6
  return(list(dz=dz,dw=dw))
}

eqmot <- function(z,w,z1,z2,qm){
  zr1 <- z - z1
  zr2 <- z - z2
  wp=-(qm*zr2/(abs(zr2))^3+zr1/(abs(zr1))^3)/(1+qm)-complex(real=0.0,imaginary=2.0)*w+z
  zp <- w
  return(list(wp=wp,zp=zp))
}
           
stream <- function(qm,rl1,nmax=250,rd=0.1){
  cm <- qm/(1+qm)
  z1 <- -cm
  z2 <- 1-cm
  wm1 <- Conj(complex(real=0.0,imaginary=-cm))
  eps <- 1e-3
  z <- complex(real=rl1-cm-eps,imaginary=0.0)
  w <- 0.0
  t <- 0.0
  dt <- 1e-4
  it <- 0
  r <- 1
  wout <- NULL
  wkout <- NULL
  while ( it < nmax ){
    it <- it + 1
    rk <- intrk(z,w,dt,z1,z2,qm)
    z <- z+rk$dz
    w <- w+rk$dw
    t <- t+dt
    if (abs(rk$dz)/abs(z) > 0.02) {dt <- dt/2}
    if (abs(rk$dz)/abs(z) < 0.005) {dt <- 2*dt}
    wi <- w + complex(real=0.0,imaginary=1.0)*z
    rold <- r
    r <- abs(z-z1)
    vk <- 1/sqrt(r*(1+qm))
    no <- Conj(z-z1)/r
    wk <- -vk*no*complex(real=0.0,imaginary=1.0)
    wk <- wk+wm1
    wout <- c(wout,wi)
    wkout <- c(wkout,Conj(wk))
  }
  return(list(vxi=Re(wout),vyi=Im(wout),vkxi=Re(wkout),vkyi=Im(wkout)))
}

vscale <- function(porb,m1,q,vfs){
  porb <- 24*3600*porb
  omega <- 2*pi/porb
  a <- (gg*m1*msun*(1+q))^(1/3)/omega^(2/3)
  vs <- omega*a/vfs
  return(vs)
}

draw.2ndlobe <- function(q,incl,porb,m1=0.6,vfs=1e5,...){
  rl1 <- rlq1(q)
  l <- lobes(q,rl1)
  cm <- q/(1+q)
  vs <- vscale(porb,m1,q,vfs)
  si <- sin(pi*incl/180.0)
  xl <- -l$y*si*vs
  yl <- (l$x-cm)*si*vs
  lines(xl,yl,...)
}

draw.stream <- function(q,incl,porb,m1=0.6,n=250,vfs=1e5,...){
  rl1 <- rlq1(q)
  st <- stream(q,rl1,nmax=n)
  vs <- vscale(porb,m1,q,vfs)
  si <- sin(pi*incl/180.0)
  lines(st$vxi*si*vs,st$vyi*si*vs,...)
}

draw.kepler <- function(q,incl,porb,m1=0.6,n=250,vfs=1e5,...){
  rl1 <- rlq1(q)
  st <- stream(q,rl1,nmax=n)
  vs <- vscale(porb,m1,q,vfs)
  si <- sin(pi*incl/180.0)
  lines(st$vkxi*si*vs,st$vkyi*si*vs,...)
}

draw.circle <- function(r,x0,y0,...){
  theta <- (1:360)*pi/180.0
  x <- r*cos(theta)+x0
  y <- r*sin(theta)+y0
  lines(x,y,...)
}

draw.center <- function(size=1,...){
  points(0,0,pch=3,cex=2*size,...)
}

draw.m2 <- function(q,incl,porb,m1=0.6,vfs=1e5,size=1,...){
  cm <- q/(1+q)
  vs <- vscale(porb,m1,q,vfs)
  si <- sin(pi*incl/180.0)
  points(0,(1-cm)*vs*si,pch=4,cex=2*size,...)
  cat(sprintf("M2: Vy=%.2f\n",(1-cm)*vs*si))
}

draw.m1 <- function(q,incl,porb,m1=0.6,vfs=1e5,size=1,...){
  cm <- q/(1+q)
  vs <- vscale(porb,m1,q,vfs)
  si <- sin(pi*incl/180.0)
  points(0,-cm*vs*si,pch=4,cex=2*size,...)
  cat(sprintf("M1: Vy=%.2f\n",-cm*vs*si))
}
